SystemOrganization addCategory: #'CS598rej-HW8'!
SystemOrganization addCategory: #'CS598rej-HW8-Test'!

ProtoObject subclass: #RetirementWithholder
	instanceVariableNames: 'percentage'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-HW8'!

!RetirementWithholder commentStamp: '<historical>' prior: 0!
I'm a withholder that wittholds a percentage of the income for the retirement fund.!

!RetirementWithholder class methodsFor: 'instance creation' stamp: 'm3r 6/27/2008 08:07'!
percentage: aPercentage
	^ self new percentage: aPercentage
! !

!RetirementWithholder methodsFor: 'initialize-release' stamp: 'm3r 6/27/2008 08:07'!
percentage: aPercentage
	percentage := aPercentage.
! !

!RetirementWithholder methodsFor: 'withholding' stamp: 'm3r 6/27/2008 08:07'!
withholdingOn: aPaycheck for: anEmployee at: aDate
	^ percentage / 100 * aPaycheck amountPaid.
! !

Object subclass: #ValueWithHistory
	instanceVariableNames: 'date value'
	classVariableNames: 'EarliestDate'
	poolDictionaries: ''
	category: 'CS598rej-HW8'!

!ValueWithHistory commentStamp: 'm3r 6/27/2008 01:58' prior: 0!
I represent a (probably numerical) value that changes over time.  I can answer my value at any point in
time (usually a date).  I can update my value at any point in time, and also add a number to my
value from any point in time on into the future.  My value does not change continuously, but changes
at discrete points in time.

Instance Variables:
	date	<SortedCollection of: Date>
	value	<OrderedCollection of: Number>
	The i'th element of value matches the i'th element of date.  The date collection indicates when the value takes on the next element of the value collection.
!

!ValueWithHistory class methodsFor: 'class initialization' stamp: 'm3r 6/27/2008 02:05'!
initialize
	"self initialize"
	EarliestDate := Date newDay: 0 year: 0.
! !

!ValueWithHistory class methodsFor: 'instance creation' stamp: 'm3r 6/27/2008 02:06'!
new
	^ super new initialize
! !

!ValueWithHistory class methodsFor: 'instance creation' stamp: 'm3r 6/27/2008 02:06'!
zero
	^ self new starting: EarliestDate become: 0
! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'm3r 6/27/2008 01:58'!
at: aDate 
	"Return value at aDate"
	| index |
	index := date indexOfStartOfIntervalContaining: aDate.
	index = 0 ifTrue: [self error: 'date is too early'].
	^value at: index
! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'm3r 6/27/2008 01:59'!
at: aDate add: amount 
"Add amount to the value starting at aDate."
	| start |
	start := self indexForAccessing: aDate.
	start to: value size do:
		[:each | value at: each put: (value at: each) + amount].
! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'm3r 6/27/2008 02:00'!
indexForAccessing: aDate 
	"Create index of slot for aDate, creating it if necessary."
	| index |
	(index := date indexOfStartOfIntervalContaining: aDate) = 0 ifTrue: 
		[date add: aDate.
		value addFirst: 0.
		^ 1].
	(date at: index) = aDate ifTrue: [^index].
	date add: aDate.
	value add: (value at: index) beforeIndex: index + 1.
	^ index + 1
! !

!ValueWithHistory methodsFor: 'initialize-release' stamp: 'MihaiCodoban 3/6/2013 21:26'!
initialize
	date := SortedSequence new.
	value := OrderedCollection new.
! !

!ValueWithHistory methodsFor: 'printing' stamp: 'm3r 6/27/2008 02:03'!
printOn: aStream 
	date with: value do: 
		[: nextDate : nextValue | 
		aStream 
			nextPut: $(;
			print: nextDate;
			nextPutAll: ' -> ';
			print: nextValue;
			nextPut: $);
			cr].
! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'm3r 6/27/2008 02:00'!
starting: aDate become: anAmount 
	| index |
	date add: aDate.
	index := date indexOfStartOfIntervalContaining: aDate.
	value add: anAmount beforeIndex: index.
! !

TestCase subclass: #HW8ExercisesTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-HW8-Test'!

!HW8ExercisesTest methodsFor: 'testing for code smells' stamp: 'm3r 6/27/2008 08:02'!
testInsuranceChange
	self should: [InsuranceChange includesSelector: #date:employee:insurance:].
	self should: [InsuranceChange includesSelector: #postTo:].
! !

!HW8ExercisesTest methodsFor: 'testing for code smells' stamp: 'm3r 6/27/2008 08:03'!
testInsuranceWithholder
	self should: [InsuranceWithholder includesSelector: #withholdingOn:for:at:].
	self should: [InsuranceWithholder selectors size = 1].
! !

!HW8ExercisesTest methodsFor: 'testing for code smells' stamp: 'm3r 6/27/2008 08:04'!
testNewPaycheck
	self shouldnt: 
		[Paycheck includesSelector: 
			#date:employee:paid:withheld:insurance:totalPaid:totalWithheld:].
	self should: [Paycheck includesSelector: #date:employee:paid:withholders:].
! !

!HW8ExercisesTest methodsFor: 'testing for code smells' stamp: 'm3r 6/27/2008 08:05'!
testNoMoreTaxesInEmployee
	| instanceVariables |
	self shouldnt: [Employee includesSelector: #taxesOn:at:].
	self shouldnt: [Employee includesSelector: #federalTaxesOn:assumingTotalIs:].
	
	instanceVariables := Employee instanceVariablesString.
	self should: [(instanceVariables findString: 'taxRule') = 0].
	self shouldnt: [(instanceVariables findString: 'withholders') = 0].
! !

!HW8ExercisesTest methodsFor: 'testing' stamp: 'm3r 6/27/2008 08:07'!
testTakeHomeAfterTaxesInsuranceRetirement
	"this method assumes that Employee gets initialized with a TaxWithholder and an 
	InsuranceWithholder. We'll just add one RetirementWittholder that withholds 10 
	percent of the income for the retirement (see class RetirementWithholder for details)"
			
	| firstDay nextDay paycheckDay john hireDate paycheck |
	
	john := Employee named: 'John Doe'.
	hireDate := Date newDay: 1 month: 5 year: 1999.
	john changeSalaryFor: hireDate to: 20.
	
	firstDay := Date newDay: 5 month: 5 year: 1999.
	john postTimeCardFor: firstDay hoursWorked: 40 vacation: 0.
	nextDay := Date newDay: 12 month: 5 year: 1999.
	john postTimeCardFor: nextDay hoursWorked: 40 vacation: 0.
	
	john changeInsuranceFor: hireDate to: 200.
	john withholders at: #retirement put: (RetirementWithholder percentage: 10).
	
	paycheckDay := Date newDay: 15 month: 5 year: 1999.
	paycheck := john makePaycheckAt: paycheckDay.
	
	self assert: paycheck amountPaid = 828.8 .
	self assert: paycheck insurance = 200.
! !

!HW8ExercisesTest methodsFor: 'testing for code smells' stamp: 'm3r 6/27/2008 08:06'!
testTaxesInTaxWithholder
	| instanceVariables classVariables |
	self should: [TaxWithholder includesSelector: #federalTaxesOn:assumingTotalIs:].
	self should: [TaxWithholder includesSelector: #withholdingOn:for:at:].
	
	instanceVariables := TaxWithholder instanceVariablesString.
	self shouldnt: [(instanceVariables findString: 'taxRule') = 0].
	
	classVariables := TaxWithholder classVariablesString findTokens: ' '.
	self should: 
		[classVariables includesAllOf: 
			#(HeadTaxRate MarriedSeparateTaxRate MarriedTaxRate SingleTaxRate)].
! !

TestCase subclass: #ValueWithHistoryTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS598rej-HW8-Test'!

!ValueWithHistoryTest methodsFor: 'testing' stamp: 'm3r 6/27/2008 02:15'!
test
	| history day3 day2 day1 day4 |
	day3 := '12/12/2007' asDate.
	day2 := day3 subtractDays: 3.
	day1 := day2 subtractDays: 3.
	day4 :=  day3 addDays: 4.
	(history := ValueWithHistory new) starting: day3 become: 3.
	history starting: day1 become: 1.
	self should: [(history at: day2) = 1].
	history at: day2 add: 1.
	self should: [(history at: day3) = 4].
	history at: day4 add: 2.
	self should: [(history at: day4) = 6].
! !
ValueWithHistory initialize!
